@@ -34,8 +34,9 @@
 #include <libsemigroups/int-range.hpp>       // for IntegralRange<>::value_type
 
 // pybind11....
-#include <pybind11/pybind11.h>  // for class_, make_iterator, init, enum_
-#include <pybind11/stl.h>       // for conversion of C++ to py types
+#include <pybind11/operators.h>  // for self, self_t, operator!=, operator*
+#include <pybind11/pybind11.h>   // for class_, make_iterator, init, enum_
+#include <pybind11/stl.h>        // for conversion of C++ to py types
 
 // libsemigroups_pybind11....
 #include "main.hpp"  // for init_action_digraph
@@ -43,9 +44,28 @@
 namespace py = pybind11;
 
 namespace libsemigroups {
+  namespace libsemigroups_pybind11 {
+    // TODO Avoid duplicate code the following is a copy of
+    // libsemigroups::action_digraph_helper::make, which can't be used because
+    // it has initializer_list's as 2nd arg.
+    ActionDigraph<size_t> make(size_t                           num_nodes,
+                               std::vector<std::vector<size_t>> il) {
+      ActionDigraph<size_t> result(num_nodes, il.begin()->size());
+      for (size_t i = 0; i < il.size(); ++i) {
+        for (size_t j = 0; j < (il.begin() + i)->size(); ++j) {
+          auto val = *((il.begin() + i)->begin() + j);
+          if (val != UNDEFINED) {
+            result.add_edge(i, *((il.begin() + i)->begin() + j), j);
+          }
+        }
+      }
+      return result;
+    }
+  }  // namespace libsemigroups_pybind11
+     //
   using node_type = ActionDigraph<size_t>::node_type;
   using algorithm = ActionDigraph<size_t>::algorithm;
-  void init_action_digraph(py::module &m) {
+  void init_action_digraph(py::module& m) {
     ////////////////////////////////////////////////////////////////////////
     // ActionDigraph
     ////////////////////////////////////////////////////////////////////////
@@ -58,17 +78,20 @@ namespace libsemigroups {
             "matrix",
             algorithm::matrix,
             R"pbdoc(Use the adjacency matrix and matrix multiplication.)pbdoc")
-        .value(
-            "acyclic",
-            algorithm::acyclic,
-            R"pbdoc(Use a dynamic programming approach for acyclic digraphs.)pbdoc")
+        .value("acyclic",
+               algorithm::acyclic,
+               R"pbdoc(
+                 Use a dynamic programming approach for acyclic digraphs.
+               )pbdoc")
         .value("trivial",
                algorithm::trivial,
                R"pbdoc(Try to utilise some corner cases.)pbdoc")
-        .value(
-            "automatic",
-            algorithm::automatic,
-            R"pbdoc(The function :py:meth:`number_of_paths` tries to decide which algorithm is best.)pbdoc");
+        .value("automatic",
+               algorithm::automatic,
+               R"pbdoc(
+                 The function :py:meth:`number_of_paths` tries to decide
+                 which algorithm is best.
+               )pbdoc");
 
     ad.def(py::init<>())
         .def(py::init<size_t>())
@@ -76,19 +99,19 @@ namespace libsemigroups {
              py::arg("m"),
              py::arg("n"),
              R"pbdoc(
-             Construct from number of nodes and out degree.
+               Construct from number of nodes and out degree.
 
-             :Parameters: - **m** (int) the number of nodes in the digraph
-                            (default: ``0``).
-                          - **n** (int) the out-degree of every node (default:
-                            ``0``).
-           )pbdoc")
-        .def(py::init<ActionDigraph<size_t> const &>(),
+               :Parameters: - **m** (int) the number of nodes in the digraph
+                              (default: ``0``).
+                            - **n** (int) the out-degree of every node (default:
+                              ``0``).
+             )pbdoc")
+        .def(py::init<ActionDigraph<size_t> const&>(),
              R"pbdoc(
                Construct a copy.
              )pbdoc")
         .def("__repr__",
-             [](ActionDigraph<size_t> const &d) {
+             [](ActionDigraph<size_t> const& d) {
                std::string result = "<action digraph with ";
                result += std::to_string(d.number_of_nodes());
                result += " nodes, ";
@@ -98,6 +121,7 @@ namespace libsemigroups {
                result += " out-degree>";
                return result;
              })
+        .def(pybind11::self == pybind11::self)
         .def("number_of_nodes",
              &ActionDigraph<size_t>::number_of_nodes,
              R"pbdoc(
@@ -105,7 +129,7 @@ namespace libsemigroups {
 
                :Parameters: None
                :return: An ``int``.
-               )pbdoc")
+             )pbdoc")
         .def("number_of_edges",
              py::overload_cast<>(&ActionDigraph<size_t>::number_of_edges,
                                  py::const_),
@@ -128,7 +152,7 @@ namespace libsemigroups {
         .def("out_degree",
              &ActionDigraph<size_t>::out_degree,
              R"pbdoc(
-           Returns the maximum out-degree of any node.
+               Returns the maximum out-degree of any node.
 
                :Parameters: None
                :return: An ``int``.
@@ -185,7 +209,8 @@ namespace libsemigroups {
              py::arg("v"),
              py::arg("lbl"),
              R"pbdoc(
-               Get the range of the edge with source node ``v`` and edge-label ``lbl``.
+               Get the range of the edge with source node ``v`` and edge-label
+               ``lbl``.
 
                :param v: the node
                :type v: int
@@ -194,7 +219,7 @@ namespace libsemigroups {
 
                :return:
                  An ``int`` or :py:obj:`UNDEFINED`.
-         )pbdoc")
+             )pbdoc")
         .def("reserve",
              &ActionDigraph<size_t>::reserve,
              py::arg("m"),
@@ -210,7 +235,7 @@ namespace libsemigroups {
                :type n: int
 
                :return: (None)
-         )pbdoc")
+             )pbdoc")
         .def("unsafe_neighbor",
              &ActionDigraph<size_t>::unsafe_neighbor,
              py::arg("v"),
@@ -225,7 +250,7 @@ namespace libsemigroups {
                :type lbl: int
 
                :return: An ``int`` or :py:obj:`UNDEFINED`.
-         )pbdoc")
+             )pbdoc")
         .def("next_neighbor",
              &ActionDigraph<size_t>::next_neighbor,
              py::arg("v"),
@@ -275,7 +300,7 @@ namespace libsemigroups {
 
                If ``neighbor(v, i)`` is undefined for every value of ``i``,
                then ``x[0]`` and ``x[1]`` equal :py:obj:`UNDEFINED`.
-         )pbdoc")
+             )pbdoc")
         .def("scc_id",
              &ActionDigraph<size_t>::scc_id,
              py::arg("nd"),
@@ -288,7 +313,7 @@ namespace libsemigroups {
 
                :Parameters: None
                :return: An ``int``.
-         )pbdoc")
+             )pbdoc")
         .def("number_of_scc",
              &ActionDigraph<size_t>::number_of_scc,
              R"pbdoc(
@@ -308,7 +333,7 @@ namespace libsemigroups {
                :type nd: int
 
                :return: An ``int``.
-         )pbdoc")
+             )pbdoc")
         .def("spanning_forest",
              &ActionDigraph<size_t>::spanning_forest,
              py::return_value_policy::copy,  // to ensure the Forest lives!
@@ -318,7 +343,7 @@ namespace libsemigroups {
                with edges oriented away from the root.
 
                :return: A :py:class:`Forest`.
-         )pbdoc")
+             )pbdoc")
         .def("reverse_spanning_forest",
              &ActionDigraph<size_t>::reverse_spanning_forest,
              R"pbdoc(
@@ -327,7 +352,7 @@ namespace libsemigroups {
                with edges oriented towards the root.
 
                :return: A :py:class:`Forest`.
-         )pbdoc")
+             )pbdoc")
         .def("number_of_paths_algorithm",
              py::overload_cast<node_type const>(
                  &ActionDigraph<size_t>::number_of_paths_algorithm, py::const_),
@@ -340,7 +365,7 @@ namespace libsemigroups {
                :Parameters: - **source** (int) the source node.
 
                :Returns: A value of type :py:obj:`ActionDigraph.algorithm`.
-         )pbdoc")
+             )pbdoc")
         .def("number_of_paths",
              py::overload_cast<node_type const>(
                  &ActionDigraph<size_t>::number_of_paths, py::const_),
@@ -360,7 +385,9 @@ namespace libsemigroups {
              py::arg("min"),
              py::arg("max"),
              R"pbdoc(
-               Returns the algorithm used by number_of_paths to compute the number of paths originating at the given source node with length in the range $[min, max)$.
+               Returns the algorithm used by number_of_paths to compute the
+               number of paths originating at the given source node with length
+               in the range $[min, max)$.
 
                :Parameters: - **source** (int) the source node
                             - **min** (int) the minimum length of paths to
@@ -368,7 +395,14 @@ namespace libsemigroups {
                             - **max** (int) the maximum length of paths to
                               count
                :Returns: A value of type :py:obj:`ActionDigraph.algorithm`.
-          )pbdoc")
+             )pbdoc")
+        .def("number_of_paths_algorithm",
+             [](ActionDigraph<size_t> const& ad,
+                node_type const              source,
+                size_t const                 min,
+                PositiveInfinity const&      max) {
+               return ad.number_of_paths_algorithm(source, min, max);
+             })
         .def("number_of_paths",
              py::overload_cast<node_type const,
                                size_t const,
@@ -416,7 +450,7 @@ namespace libsemigroups {
                             - **max**    (int) the maximum length of paths to count
 
                :Returns: A value of type :py:obj:`ActionDigraph.algorithm`.
-         )pbdoc")
+             )pbdoc")
         .def("number_of_paths",
              py::overload_cast<node_type const,
                                node_type const,
@@ -442,10 +476,10 @@ namespace libsemigroups {
                               :py:obj:`ActionDigraph.algorithm.automatic`)
 
                :Returns: An ``int``.
-         )pbdoc")
+             )pbdoc")
         .def(
             "number_of_paths",
-            [](ActionDigraph<size_t> const &ad,
+            [](ActionDigraph<size_t> const& ad,
                node_type                    source,
                node_type                    target,
                size_t                       min,
@@ -457,19 +491,19 @@ namespace libsemigroups {
             py::arg("min"),
             py::arg("max"),
             R"pbdoc(
-               Returns the number of paths between a pair of nodes with length
-               in a given range.
+              Returns the number of paths between a pair of nodes with length
+              in a given range.
 
-               :Parameters: - **source** (int) the source node
-                            - **target** (int) the target node
-                            - **min**    (int) the minimum length of paths to count
-                            - **max**    (int) the maximum length of paths to count
+              :Parameters: - **source** (int) the source node
+                           - **target** (int) the target node
+                           - **min**    (int) the minimum length of paths to count
+                           - **max**    (int) the maximum length of paths to count
 
-               :Returns: An ``int``.
-         )pbdoc")
+              :Returns: An ``int``.
+            )pbdoc")
         .def(
             "number_of_paths",
-            [](ActionDigraph<size_t> const &ad,
+            [](ActionDigraph<size_t> const& ad,
                node_type                    source,
                size_t                       min,
                size_t max) { return ad.number_of_paths(source, min, max); },
@@ -477,35 +511,41 @@ namespace libsemigroups {
             py::arg("min"),
             py::arg("max"),
             R"pbdoc(
-               Returns the number of paths starting at a node with length
-               in a given range.
+              Returns the number of paths starting at a node with length
+              in a given range.
 
-               :Parameters: - **source** (int) the source node
-                            - **min**    (int) the minimum length of paths to count
-                            - **max**    (int) the maximum length of paths to count
+              :Parameters: - **source** (int) the source node
+                           - **min**    (int) the minimum length of paths to count
+                           - **max**    (int) the maximum length of paths to count
 
-               :Returns: An ``int``.
-         )pbdoc")
+              :Returns: An ``int``.
+            )pbdoc")
+        .def("number_of_paths",
+             [](ActionDigraph<size_t> const& ad,
+                node_type                    source,
+                size_t                       min,
+                PositiveInfinity             max) {
+               return ad.number_of_paths(source, min, max);
+             })
         .def(
             "nodes_iterator",
-            [](ActionDigraph<size_t> const &ad) {
+            [](ActionDigraph<size_t> const& ad) {
               return py::make_iterator(ad.cbegin_nodes(), ad.cend_nodes());
             },
             R"pbdoc(
-          Returns an iterator to the nodes of the digraph.
+              Returns an iterator to the nodes of the digraph.
             )pbdoc")
         .def(
             "reverse_nodes_iterator",
-            [](ActionDigraph<size_t> const &ad) {
+            [](ActionDigraph<size_t> const& ad) {
               return py::make_iterator(ad.crbegin_nodes(), ad.crend_nodes());
             },
             R"pbdoc(
                 Returns a reversed iterator to the nodes of the digraph.
-    )pbdoc")
-
+            )pbdoc")
         .def(
             "edges_iterator",
-            [](ActionDigraph<size_t> const &ad, size_t const i) {
+            [](ActionDigraph<size_t> const& ad, size_t const i) {
               return py::make_iterator(ad.cbegin_edges(i), ad.cend_edges(i));
             },
             py::arg("i"),
@@ -519,7 +559,7 @@ namespace libsemigroups {
             )pbdoc")
         .def(
             "sccs_iterator",
-            [](ActionDigraph<size_t> const &ad) {
+            [](ActionDigraph<size_t> const& ad) {
               return py::make_iterator(ad.cbegin_sccs(), ad.cend_sccs());
             },
             R"pbdoc(
@@ -527,16 +567,16 @@ namespace libsemigroups {
             )pbdoc")
         .def(
             "scc_iterator",
-            [](ActionDigraph<size_t> const &ad, size_t const i) {
+            [](ActionDigraph<size_t> const& ad, size_t const i) {
               return py::make_iterator(ad.cbegin_scc(i), ad.cend_scc(i));
             },
             R"pbdoc(
-        Returns an iterator pointing to the first node in the scc with the specified id-number.
-    )pbdoc")
-
+              Returns an iterator pointing to the first node in the scc with the
+              specified id-number.
+            )pbdoc")
         .def(
             "scc_roots_iterator",
-            [](ActionDigraph<size_t> const &ad) {
+            [](ActionDigraph<size_t> const& ad) {
               return py::make_iterator(ad.cbegin_scc_roots(),
                                        ad.cend_scc_roots());
             },
@@ -545,10 +585,10 @@ namespace libsemigroups {
             )pbdoc")
         .def(
             "panilo_iterator",
-            [](ActionDigraph<size_t> const &ad,
-               node_type const &            source,
-               size_t const &               mn,
-               size_t const &               mx) {
+            [](ActionDigraph<size_t> const& ad,
+               node_type const&             source,
+               size_t const&                mn,
+               size_t const&                mx) {
               return py::make_iterator(ad.cbegin_panilo(source, mn, mx),
                                        ad.cend_panilo());
             },
@@ -572,12 +612,20 @@ namespace libsemigroups {
 
               :return: An iterator.
             )pbdoc")
+        .def("panilo_iterator",
+             [](ActionDigraph<size_t> const& ad,
+                node_type const&             source,
+                size_t const&                mn,
+                PositiveInfinity const&      mx) {
+               return py::make_iterator(ad.cbegin_panilo(source, mn, mx),
+                                        ad.cend_panilo());
+             })
         .def(
             "panislo_iterator",
-            [](ActionDigraph<size_t> const &ad,
-               node_type const &            source,
-               size_t const &               mn,
-               size_t const &               mx) {
+            [](ActionDigraph<size_t> const& ad,
+               node_type const&             source,
+               size_t const&                mn,
+               size_t const&                mx) {
               return py::make_iterator(ad.cbegin_panislo(source, mn, mx),
                                        ad.cend_panislo());
             },
@@ -601,12 +649,20 @@ namespace libsemigroups {
               :type max: int
               :return: An iterator.
             )pbdoc")
+        .def("panislo_iterator",
+             [](ActionDigraph<size_t> const& ad,
+                node_type const&             source,
+                size_t const&                mn,
+                PositiveInfinity const&      mx) {
+               return py::make_iterator(ad.cbegin_panislo(source, mn, mx),
+                                        ad.cend_panislo());
+             })
         .def(
             "pilo_iterator",
-            [](ActionDigraph<size_t> const &ad,
-               node_type const &            source,
-               size_t const &               mn,
-               size_t const &               mx) {
+            [](ActionDigraph<size_t> const& ad,
+               node_type const&             source,
+               size_t const&                mn,
+               size_t const&                mx) {
               return py::make_iterator(ad.cbegin_pilo(source, mn, mx),
                                        ad.cend_pilo());
             },
@@ -629,12 +685,20 @@ namespace libsemigroups {
 
               :return: An iterator.
             )pbdoc")
+        .def("pilo_iterator",
+             [](ActionDigraph<size_t> const& ad,
+                node_type const&             source,
+                size_t const&                mn,
+                PositiveInfinity const&      mx) {
+               return py::make_iterator(ad.cbegin_pilo(source, mn, mx),
+                                        ad.cend_pilo());
+             })
         .def(
             "pislo_iterator",
-            [](ActionDigraph<size_t> const &ad,
-               node_type const &            source,
-               size_t const &               mn,
-               size_t const &               mx) {
+            [](ActionDigraph<size_t> const& ad,
+               node_type const&             source,
+               size_t const&                mn,
+               size_t const&                mx) {
               return py::make_iterator(ad.cbegin_pislo(source, mn, mx),
                                        ad.cend_pislo());
             },
@@ -659,13 +723,21 @@ namespace libsemigroups {
 
               :return: An iterator.
             )pbdoc")
+        .def("pislo_iterator",
+             [](ActionDigraph<size_t> const& ad,
+                node_type const&             source,
+                size_t const&                mn,
+                PositiveInfinity const&      mx) {
+               return py::make_iterator(ad.cbegin_pislo(source, mn, mx),
+                                        ad.cend_pislo());
+             })
         .def(
             "pstislo_iterator",
-            [](ActionDigraph<size_t> const &ad,
-               node_type const &            source,
-               node_type const &            target,
-               size_t const &               mn,
-               size_t const &               mx) {
+            [](ActionDigraph<size_t> const& ad,
+               node_type const&             source,
+               node_type const&             target,
+               size_t const&                mn,
+               size_t const&                mx) {
               return py::make_iterator(
                   ad.cbegin_pstislo(source, target, mn, mx), ad.cend_pstislo());
             },
@@ -693,6 +765,16 @@ namespace libsemigroups {
 
               :return: An iterator.
             )pbdoc")
+        .def("pstislo_iterator",
+             [](ActionDigraph<size_t> const& ad,
+                node_type const&             source,
+                node_type const&             target,
+                size_t const&                mn,
+                PositiveInfinity const&      mx) {
+               return py::make_iterator(
+                   ad.cbegin_pstislo(source, target, mn, mx),
+                   ad.cend_pstislo());
+             })
         .def_static(
             "random",
             [](size_t nr_nodes, size_t out_degree) {
@@ -739,7 +821,7 @@ namespace libsemigroups {
     // can be included here!
 
     m.def("add_cycle",
-          py::overload_cast<ActionDigraph<size_t> &, size_t>(
+          py::overload_cast<ActionDigraph<size_t>&, size_t>(
               &action_digraph_helper::add_cycle<size_t>),
           py::arg("ad"),
           py::arg("N"),
@@ -754,21 +836,21 @@ namespace libsemigroups {
             :return: None.
           )pbdoc");
     m.def("is_acyclic",
-          py::overload_cast<ActionDigraph<size_t> const &>(
+          py::overload_cast<ActionDigraph<size_t> const&>(
               &action_digraph_helper::is_acyclic<size_t>),
           py::arg("ad"),
           R"pbdoc(
-          Check if a digraph is acyclic.
+            Check if a digraph is acyclic.
 
-          A digraph is acyclic if every directed cycle is trivial.
+            A digraph is acyclic if every directed cycle is trivial.
 
-          :param ad: the digraph.
-          :type ad: ActionDigraph
+            :param ad: the digraph.
+            :type ad: ActionDigraph
 
-          :return: A ``bool``.
+            :return: A ``bool``.
           )pbdoc");
     m.def("topological_sort",
-          py::overload_cast<ActionDigraph<size_t> const &>(
+          py::overload_cast<ActionDigraph<size_t> const&>(
               &action_digraph_helper::topological_sort<size_t>),
           py::arg("ad"),
           R"pbdoc(
@@ -783,7 +865,7 @@ namespace libsemigroups {
             :Returns: A list of ``int``.
           )pbdoc");
     m.def("topological_sort",
-          py::overload_cast<ActionDigraph<size_t> const &, size_t>(
+          py::overload_cast<ActionDigraph<size_t> const&, size_t>(
               &action_digraph_helper::topological_sort<size_t>),
           py::arg("ad"),
           py::arg("source"),
@@ -822,5 +904,8 @@ namespace libsemigroups {
               if ``source`` is not a node in the digraph or ``path`` contains a
               value that is not an edge-label.
           )pbdoc");
+    m.def("action_digraph_helper_make",
+          &libsemigroups_pybind11::make,
+          R"pbdoc( TODO )pbdoc");
   }
 }  // namespace libsemigroups
